module ML.AST where

import List
import Data.Generics
import Data.Typeable

import Phrac.Pretty
import Phrac.Error
import Phrac.AbstractSyntax hiding ( ValId )

import ML.Identifiers

--
-- Modules language
--

data SigBod = OpaqueTyp TypId
            | ManifestTyp TypId SimTyp
            | ValSpec ValId ValTyp
              deriving (Show,Eq,Typeable,Data)

data SigExp = SigBod [SigBod]
            | SigVar SigId
            | SigWhere SigExp TypId SimTyp
              deriving (Show,Eq,Typeable,Data)

data StrBod = TypDef TypId SimTyp
            | ValDef ValId (Maybe ValTyp) ValExp
              deriving (Show,Eq,Typeable,Data)

data StrExp = StrBod [StrBod]
            | StrVar StrId
            | FunctorApp FunId [StrExp]
            | SealedStrExp StrExp SigExp
              deriving (Show,Eq,Typeable,Data)

data Rfun = Rfun FunId [(StrId, SigExp)] SigExp StrExp
            deriving (Show,Eq,Typeable,Data)

data Rfuns = Rfuns [Rfun]
             deriving (Show,Eq,Typeable,Data)

data SigBind = SigBind SigId SigExp
               deriving (Show,Eq,Typeable,Data)

data ToplevelVal = ToplevelVal ValId ValExp
                   deriving (Show,Eq,Typeable,Data)

data Prog = Prog [ToplevelVal] [SigBind] Rfuns ValExp
            deriving (Show,Eq,Typeable,Data)

--
-- Core language
--

-- Types
--
data SimTyp = SimTypTyvar SimTyvar          -- 'a
            | SimTypTycon TyconId [SimTyp]  -- T u1 ... uN
            | TmpTyApp SimTyp SimTyp        -- only used during parsing
            | ModTyp TypId                  -- t
            | QModTyp StrId TypId           -- X.t
            | PackedStr SigExp
              deriving (Show,Eq,Typeable,Data)

data ValTyp = ValTyp [SimTyvar] SimTyp
              deriving (Show,Eq,Typeable,Data)

isTyvar (SimTypTyvar _) = True
isTyvar _         = False

freeTyvars :: SimTyp -> [SimTyvar]
freeTyvars (SimTypTyvar a) = [a]
freeTyvars (SimTypTycon _ l) = foldr union [] (map freeTyvars l)
freeTyvars _ = []

-- Expressions
--
data Binop = Add | Sub | Mult | Div | And | Or
           | Eq | Neq | Leq | Lt | Geq | Gt
             deriving (Show,Eq,Typeable,Data)

data ValExp = IntExp Integer              -- 42
            | BinExp Binop ValExp ValExp  -- 1 + 2
            | CoreVarExp CoreId           -- c
            | LamExp CoreId ValExp        -- \c -> e
            | AppExp ValExp ValExp        -- e e
            | LetExp CoreId (Maybe ValTyp) ValExp ValExp -- let c = e in e
            | IfExp ValExp ValExp ValExp  -- if e then e else e
            | ModVarExp ValId             -- x
            | QModVarExp StrId ValId      -- X.x
            | PackExp StrExp SigExp       -- pack s as S
            | OpenExp ValExp StrId SigExp ValExp  -- open e as X:S in e
            | ListExp [ValExp]            -- [e1,e2,...]
            | NilExp
            | ConsExp ValExp ValExp
            | CodeExp String              -- arbitrary pieces of code !!HACK
            | OverloadingML Overloading
              deriving (Show,Eq,Typeable,Data)

isFunctional (LamExp _ _) = True
isFunctional _            = False

--
-- Pretty printing
--

recVar = "R___"        -- recursion variable. used as ``rec(R___ : ...)''

instance Pretty SigBod where
    ppr (OpaqueTyp t)     = text "type" <+> ppr t
    ppr (ManifestTyp t u) = text "type" <+> ppr t <=> ppr u
    ppr (ValSpec x v)     = text "val" <+> ppr x <:> ppr v
    pprTML (OpaqueTyp t)     = text "type" <+> pprTML t
    pprTML (ManifestTyp t u) = text "type" <+> pprTML t <=> pprTML u
    pprTML (ValSpec x v)     = text "val" <+> pprTML x <:> pprTML v

instance Pretty SigExp where
    ppr (SigBod [])      = text "sig end"
    ppr (SigBod bod)     = text "sig" $$ nest 2 (vcat (map ppr bod)) $$
                           text "end"
    ppr (SigVar v)       = ppr v
    ppr (SigWhere s t u) = ppr s <+> text "where type" <+> ppr t <=> ppr u
    pprTML (SigBod [])      = text "sig end"
    pprTML (SigBod bod)     = text "sig" $$ nest 2 (vcat (map pprTML bod)) $$
                           text "end"
    pprTML (SigVar v)       = pprTML v
    pprTML (SigWhere s t u) = pprTML s <+> text "where type" <+> pprTML t <=> pprTML u

instance Pretty StrBod where
    ppr (TypDef t u)           = text "type" <+> ppr t <=> ppr u
    ppr (ValDef x Nothing e)   = text "val" <+> ppr x <=> ppr e
    ppr (ValDef x (Just (ValTyp vars u)) e) = 
        text "val" <+> (hsep $ map ppr vars) <+> ppr x <:> ppr u <=> ppr e
    pprTML (TypDef t u)           = text "type" <+> pprTML t <=> pprTML u
    pprTML (ValDef x Nothing e)   = text "val" <+> pprTML x <=> pprTML e
    pprTML (ValDef x (Just (ValTyp vars u)) e) = 
        text "val" <+> (hsep $ map pprTML vars) <+> pprTML x <:> pprTML u <=> pprTML e
 
instance Pretty StrExp where
    ppr (StrVar x)             = ppr x
    ppr (StrBod [])            = text "struct end"
    ppr (StrBod bod)           = text "struct" $$ nest 2 (vcat (map ppr bod))
                                 $$ text "end"
    ppr (FunctorApp f xs)      = text recVar <.> ppr f <+> 
                                 hsep (map (\ a -> parens $ ppr a) xs)
    ppr (SealedStrExp str sig) = ppr str <+> text ":>" <+> ppr sig
    pprTML (StrVar x)             = pprTML x
    pprTML (StrBod [])            = text "struct end"
    pprTML (StrBod bod)           = text "struct" $$ nest 2 (vcat (map pprTML bod))
                                 $$ text "end"
    pprTML (FunctorApp f xs)      = text recVar <.> pprTML f <+> 
                                 hsep (map (\ a -> parens $ pprTML a) xs)
    pprTML (SealedStrExp str sig) = pprTML str <+> text ":>" <+> pprTML sig

instance Pretty Rfun where
    ppr (Rfun f args _ bod) = 
        text "functor" <+> ppr f <+>
        hsep (map (\ (a,b) -> parens $ ppr a <:> ppr b) args) <=> ppr bod
    pprTML (Rfun f args sig bod) = 
        text "functor" <+> pprTML f <+>
        hsep (map (\ (a,b) -> parens $ pprTML a <:> pprTML b) args)
        <:> pprTML sig <=> pprTML bod

instance Pretty Rfuns where
    ppr (Rfuns l) = 
        text "structure" <+> text recVar <=> text "rec" <+>
        parens (text recVar <:> text "sig" <+> 
                (vcat $ map pprRfunSig l) <+> text "end") $$
        (nest 2 $ text "struct" $$ (nest 2 $ vcat $ map ppr l) $$ text "end")
        where pprRfunSig (Rfun f args sig _) = 
                  text "functor" <+> ppr f <:>
                  (hsep $ punctuate (text "->") (map pprRfunArgSig args ++ [ppr sig]))
              pprRfunArgSig (x, sig) = 
                  text "functor" <+> parens (ppr x <:> ppr sig)
    pprTML (Rfuns l) = 
        text "rec" $$ nest 2 (vcat $ map pprTML l)

instance Pretty SigBind where
    ppr (SigBind id sig) = text "signature" <+> ppr id <=> ppr sig
    pprTML (SigBind id sig) = text "signature" <+> pprTML id <=> pprTML sig
        
instance Pretty ToplevelVal where
    ppr (ToplevelVal id e) = text "val" <+> ppr id <=> ppr e
    pprTML (ToplevelVal id e) = text "val" <+> pprTML id <=> pprTML e

instance Pretty Prog where
    ppr (Prog vals sigs rfuns exp) = 
        vcat (map ppr vals) $$ vcat (map ppr sigs) $$ ppr rfuns $$
        text "val _ = print (Int.toString" <+> parens (ppr exp) <+> text ");"
    pprTML (Prog vals sigs rfuns exp) = 
        vcat (map pprTML vals) $$ vcat (map pprTML sigs) $$ pprTML rfuns $$
        text "structure Main = struct val main =" 
        <+> pprTML exp <+> text "end"

instance Pretty SimTyp where
    ppr (SimTypTyvar a)      = ppr a
    ppr (SimTypTycon (TyconId "->") [t1,t2]) 
        = parens (ppr t1 <+> text "->" <+> ppr t2)
    ppr (SimTypTycon (TyconId "->") l) 
        = parens $ text "(->)" <+> hcat (map ppr l)
    ppr (SimTypTycon n ts)   = parens (vcat (map ppr ts) <+> ppr n)
    ppr (ModTyp t)           = ppr t
    ppr (QModTyp x t)        = ppr x <.> ppr t
    ppr (PackedStr sig)      = text "[" <+> ppr sig <+> text "]"
    ppr (TmpTyApp u1 u2)     = parens (ppr u1) <> text "@" <> parens (ppr u2)
    pprTML (SimTypTyvar a)      = pprTML a
    pprTML (SimTypTycon (TyconId "->") [t1,t2]) 
        = parens (pprTML t1 <+> text "->" <+> pprTML t2)
    pprTML (SimTypTycon (TyconId "->") l) 
        = parens $ text "(->)" <+> hcat (map ppr l)
    pprTML (SimTypTycon n ts)   = parens (pprTML n <+> vcat (map pprTML ts))
    pprTML (ModTyp t)           = pprTML t
    pprTML (QModTyp x t)        = pprTML x <.> pprTML t
    pprTML (PackedStr sig)      = text "[" <+> pprTML sig <+> text "]"

instance Pretty ValTyp where
    ppr (ValTyp _ u) = ppr u
    pprTML (ValTyp _ u) = pprTML u

instance Pretty Binop where
    ppr Add   = text "+"
    ppr Sub   = text "-"
    ppr Mult  = text "*"
    ppr Div   = text "/"
    ppr And   = text "andalso"
    ppr Or    = text "orelse"
    ppr Eq    = text "="
    ppr Neq   = text "<>"
    ppr Leq   = text "<="
    ppr Lt    = text "<"
    ppr Geq   = text ">="
    ppr Gt    = text ">"
--    ppr op    = error ("No pretty printer for binary operator " ++ show op)
    pprTML Add   = text "+"
    pprTML Sub   = text "-"
    pprTML Mult  = text "*"
    pprTML Div   = text "/"
    pprTML And   = text "&&"
    pprTML Or    = text "||"
    pprTML Eq    = text "=="
    pprTML Neq   = text "/="
    pprTML Leq   = text "<="
    pprTML Lt    = text "<"
    pprTML Geq   = text ">="
    pprTML Gt    = text ">"
--    pprTML op    = error ("No pretty printer for binary operator " ++ show op)
    
instance Pretty ValExp where
    ppr (IntExp i)          = integer i
    ppr (BinExp op e1 e2)   = parens (ppr e1 <+> ppr op <+> ppr e2)
    ppr (CoreVarExp c)      = ppr c
    ppr (LamExp c e)        = parens (text "fn" <+> ppr c <+> text "=>"
                                      <+> ppr e)
    ppr (AppExp e1 e2)      = parens (ppr e1 <+> ppr e2)
    ppr (LetExp c Nothing e1 e2) = 
        parens (text "let val" <+> ppr c <=> ppr e1
                <+> text "in" <+> ppr e2 <+> text "end")
    ppr (LetExp c (Just (ValTyp vars u)) e1 e2) = 
        parens (text "let val" <+> (hsep $ map ppr vars) <+> 
                ppr c <:> ppr u <=> ppr e1
                <+> text "in" <+> ppr e2 <+> text "end")
    ppr (IfExp e1 e2 e3)    = parens (text "if" <+> ppr e1 <+> text "then" <+>
                                      ppr e2 <+> text "else" <+> ppr e3)
    ppr (ModVarExp x)       = ppr x
    ppr (QModVarExp x y)    = ppr x <> char '.' <> ppr y
    ppr (PackExp str sig)   = text "[ structure " <> ppr str <> text " as " <>
                              ppr sig <> text " ]"
    ppr (OpenExp e1 x sig e2) = 
        text "let structure " <> ppr x <> text " as " <> ppr sig <=> ppr e1 <+>
        text "in" <+> ppr e2 <+> text "end"
    ppr (ListExp es) = 
        text "[" <> (hcat $ punctuate (text "; ") (map ppr es)) <> text "]"
    ppr NilExp = text "[]"
    ppr (ConsExp e1 e2) = parens (parens (ppr e1) <+> text "::" <+>
                                  parens (ppr e2))
    ppr (CodeExp s) = text s
    ppr e = panic ("cannot pretty print " ++ show e)
    pprTML (IntExp i)          = integer i
    pprTML (BinExp op e1 e2)   = parens (pprTML e1 <+> pprTML op <+> pprTML e2)
    pprTML (CoreVarExp c)      = pprTML c
    pprTML (LamExp c e)        = parens (text "\\" <+> pprTML c <+> text "."
                                      <+> pprTML e)
    pprTML (AppExp e1 e2)      = parens (pprTML e1 <+> pprTML e2)
    pprTML (LetExp c Nothing e1 e2) = 
        parens (text "let" <+> pprTML c <=> pprTML e1
                <+> text "in" <+> pprTML e2)
    pprTML (LetExp c (Just (ValTyp vars u)) e1 e2) = 
        parens (text "let" <+> (hsep $ map pprTML vars) <+> 
                pprTML c <:> text "forall {" <> hsep (map pprTML vars) <> text "} . "
                <> pprTML u <=> pprTML e1
                <+> text "in" <+> pprTML e2)
    pprTML (IfExp e1 e2 e3)    = parens (text "if" <+> pprTML e1 <+> text "then" <+>
                                      pprTML e2 <+> text "else" <+> pprTML e3)
    pprTML (ModVarExp x)       = pprTML x
    pprTML (QModVarExp x y)    = pprTML x <> char '.' <> pprTML y
    pprTML (PackExp str sig)   = text "pack" <+> pprTML str <+> text "as" <+> 
                                 pprTML sig
    pprTML (OpenExp e1 x sig e2) = 
        text "open" <+> pprTML e1 <+> text "as" <+> pprTML x <:> pprTML sig <+>
        text "in" <+> pprTML e2
    pprTML (ListExp es) = 
        text "[" <> (hcat $ punctuate (text ", ") (map pprTML es)) <> text "]"
    pprTML NilExp = text "[]"
    pprTML (ConsExp e1 e2) = parens (parens (pprTML e1) <+> text "::" <+>
                                     parens (pprTML e2))
    pprTML (CodeExp s) = text s
    pprTML e = panic ("cannot pretty print " ++ show e)
  
--
-- Utilities
--

tyFun :: [SimTyp] -> SimTyp
tyFun [t1,t2] = SimTypTycon (TyconId "->") [t1,t2]
tyFun (t:ts)  = SimTypTycon (TyconId "->") [t, tyFun ts]
tyFun _       = panic "not enough types to construct a function type"

instance Monad (Either String) where
    Left s  >>= _ = Left s
    Right x >>= f = f x
    return x      = Right x
    fail s        = Left s

removeTmpTyApps :: (Pretty a, Data a, Show a) => a -> a
removeTmpTyApps x = 
    case removeTmpTyApps' x of
      Right y -> y
      Left err -> 
          panic ("removeTmpTyApps failed for input " ++ show x ++ ": " ++ err)

removeTmpTyApps' :: (Monad m, Data a) => a -> m a
removeTmpTyApps' x = everywhereM (mkM flattenTmpTyApps) x
    where flattenTmpTyApps :: Monad m => SimTyp -> m SimTyp
          flattenTmpTyApps t@(TmpTyApp _ _) = flatten t
          flattenTmpTyApps x = return x
          flatten (TmpTyApp (SimTypTycon id []) t) = return $ SimTypTycon id [t]
          flatten (TmpTyApp t1 t2) = 
              do t1' <- flatten t1
                 case t1' of
                   SimTypTycon id l -> return $ SimTypTycon id (l++[t2])
                   _ -> fail "error flattening temporary type application"
          flatten u = return u

