module ML.TMLParser (parseProg, parseString) where

import List
import Char
import Maybe
import qualified Data.Map as Map

import Text.ParserCombinators.Parsec
import qualified Text.ParserCombinators.Parsec.Token as T
import Text.ParserCombinators.Parsec.Language
import Text.ParserCombinators.Parsec.Expr

import ML.AST
import ML.Identifiers

data MLState = MLState
    { valueIdentifiers :: [ValId]
    }

emptyMLState = MLState { valueIdentifiers = [] }

type MLParser a = CharParser MLState a

--
-- Lexer
--

lexer :: T.TokenParser MLState
lexer = T.makeTokenParser 
        (haskellStyle
         { reservedNames = ["type", "val", "sig", "struct", "end", "where",
                            "open", "as", "pack", "forall",
                            "structure", "functor", "signature",
                            "let", "in", "if", "then", "else"]
         , reservedOpNames = ["=", "\\", "->", ":>",
                              -- boolean operators
                              "&&", "||", 
                              -- comparison operators
                              "==", "/=", "<", ">", "<=", ">=",
                              -- arithmetic operators
                              "+", "-", "*", "/"
                             ]
         , identStart = (letter <|> char '\'')
         , identLetter = (alphaNum <|> char '.' <|> char '\'')
         , commentStart = "(*"
         , commentEnd = "*)"
         , nestedComments = True
         })

whiteSpace = T.whiteSpace lexer
lexeme     = T.lexeme lexer
symbol     = T.symbol lexer
natural    = T.natural lexer
parens     = T.parens lexer
brackets   = T.brackets lexer
semi       = T.semi lexer
identifier = T.identifier lexer
reserved   = T.reserved lexer
reservedOp = T.reservedOp lexer
commaSep1  = T.commaSep1 lexer
semiSep    = T.semiSep lexer
commaSep   = T.commaSep lexer
colon      = T.colon lexer
dot        = T.dot lexer
braces     = T.braces lexer
squares    = T.squares lexer

splitAtDot :: String -> MLParser (Either String (String,String))
splitAtDot s =
    case findIndices ((==) '.') s of
      [i] -> case splitAt i s of
               ([], _) -> fail ("invalid qualified identifier")
               (_, ['.']) -> fail ("invalid qualified identifier")
               (mod, '.':val) -> return $ Right (mod, val)
      [] -> return $ Left s
      _ -> fail ("invalid qualified identifier")

--
-- Core expression parser
--

optionalTypeAnnotation :: MLParser (Maybe ValTyp)
optionalTypeAnnotation =
    option Nothing (do colon
                       v <- valtyp
                       return $ Just $ v)

parseIdentifier :: String -> MLParser ValExp
parseIdentifier ('\'' : _) = 
    fail ("type variable used instead of an identifier")
parseIdentifier s =
    do x <- splitAtDot s
       case x of
         Right (mod, val) -> return $ QModVarExp (StrId mod) (ValId val)
         Left val ->
            do state <- getState
               if (ValId val) `elem` valueIdentifiers state
                  then return $ ModVarExp (ValId val)
                  else return $ CoreVarExp (CoreId val)

assertCoreId :: String -> MLParser ()    
assertCoreId (x:_) | isLower x = return ()
assertCoreId s@(_) = fail ("invalid core identifier: " ++ show s)

binopExp :: MLParser ValExp
binopExp = buildExpressionParser table factor <?> "binary expression"
    where
    table :: OperatorTable Char MLState ValExp
    table = [ [Infix (return AppExp <?> "function application") AssocLeft]
            , [op "*" Mult AssocLeft, op "/" Div AssocLeft]
            , [op "+" Add AssocLeft, op "-" Sub AssocLeft]
            , [op "==" Eq AssocNone, op "/=" Neq AssocNone,
               op "<=" Leq AssocNone, op "<" Lt AssocNone,
               op ">=" Geq AssocNone, op ">" Gt AssocNone]
            , [op "&&" And AssocLeft]
            , [op "||" Or AssocLeft]]
    op :: String -> Binop -> Assoc -> Operator Char MLState ValExp
    op s o assoc = Infix ((do reservedOp s
                              return $ BinExp o)
                          <?> "binary operator")
                         assoc
    factor = parens valexp
         <|> do i <- natural
                return $ IntExp i
         <|> do x <- identifier 
                parseIdentifier x
         <|> do es <- squares (commaSep valexp)
                return $ ListExp es


valexp :: MLParser ValExp
valexp = binopExp
     <|> do reservedOp "\\"
            x <- identifier
            assertCoreId x
            reservedOp "."
            e <- valexp
            return $ LamExp (CoreId x) e
     <|> do reserved "let"
            x <- identifier
            assertCoreId x
            v <- optionalTypeAnnotation
            reservedOp "="
            e1 <- valexp
            reserved "in"
            e2 <- valexp
            return $ LetExp (CoreId x) v e1 e2
     <|> do reserved "if"
            e1 <- valexp
            reserved "then"
            e2 <- valexp
            reserved "else"
            e3 <- valexp
            return $ IfExp e1 e2 e3
     <|> do reserved "open"
            e <- valexp
            reserved "as"
            x <- coreIdentifier
            colon
            s <- sigexp
            reserved "in"
            e' <- valexp
            return $ OpenExp e (StrId x) s e'
     <|> do reserved "pack"
            str <- sealedStrexp
            reserved "as"
            sig <- sigexp
            return $ PackExp str sig

--
-- Type parser
--

tycons = ["int", "list"]

simtyvar :: MLParser SimTyvar
simtyvar = 
    do x <- identifier
       case x of
         '\'' : _ -> return (SimTyvar x)
         _ -> fail ("expected type variable, not `" ++ x ++ "'")

parseTypeIdentifier :: String -> MLParser SimTyp
parseTypeIdentifier s@('\'':_) = return $ SimTypTyvar (SimTyvar s)
parseTypeIdentifier s =
    do x <- splitAtDot s
       case x of
         Right (mod, typ) -> return $ QModTyp (StrId mod) (TypId typ)
         Left typ ->
             if typ `elem` tycons
                then return $ SimTypTycon (TyconId typ) []
                else return $ ModTyp (TypId typ)

simtyp = buildExpressionParser table factor <?> "simple type"
    where
    table :: OperatorTable Char MLState SimTyp
    table = [ [Infix (return TmpTyApp <?> "function application") AssocLeft]
            , [Infix ((do reservedOp "->"
                          return $ \t1 t2 -> SimTypTycon (TyconId "->") [t1,t2])
                      <?> "function space type constructor")
                    AssocRight]]
    factor = parens simtyp
         <|> do x <- identifier
                parseTypeIdentifier x
         <|> do reservedOp "<"
                s <- sigexp
                reservedOp ">"
                return $ PackedStr s

tyvar = 
    do char '\''
       x <- identifier
       return ('\'' : x)
     
valtyp =
    do as <- option [] (do reserved "forall"
                           as <- braces (commaSep simtyvar)
                           dot
                           return as)
       u <- simtyp
       return $ ValTyp as u
  
--
-- Structure expression parser
--

coreIdentifier = 
    do x <- identifier
       x' <- parseIdentifier x
       case x' of 
         CoreVarExp (CoreId x'') -> return x''
         _ -> fail ("invalid identifiers: " ++ show x)

typeIdentifier =
    do t <- identifier
       t' <- parseTypeIdentifier t
       case t' of
         ModTyp t'' -> return t''
         _ -> fail ("invalid type identifier: " ++ show t)

strbodElem = 
      do reserved "type"
         t <- typeIdentifier
         reservedOp "="
         u <- simtyp
         return $ TypDef t u
  <|> do reserved "val"
         x <- coreIdentifier
         v <- optionalTypeAnnotation
         reservedOp "="
         -- we register x as a value identifier BEFORE parsing the 
         -- expression in order to allow recursive bindings. This may lead
         -- to problems if the expression is not functional and x is 
         -- already in scope. I don't care about these problems.
         state <- getState
         setState (state { valueIdentifiers = 
                           (ValId x) : valueIdentifiers state })
         e <- valexp
         return $ ValDef (ValId x) v e

primStrexp = 
    do reserved "struct"
       updateState (\state -> state { valueIdentifiers = [] })
       elems <- many strbodElem
       updateState (\state -> state { valueIdentifiers = [] })
       reserved "end"
       return $ StrBod elems

strexp = 
        try (do x <- coreIdentifier
                args <- (parens $ commaSep strexp)
                let args' = if null args 
                               then [StrBod []]
                               else args
                return $ FunctorApp (FunId x) args')
    <|> do x <- coreIdentifier
           return $ StrVar (StrId x) 
    <|> primStrexp

sealedStrexp = 
    do s <- strexp
       option s
              (do reservedOp ":>"
                  sig <- sigexp
                  return $ SealedStrExp s sig)


--
-- Signature expression parser
--

sigbodElem =
      try (do reserved "type"
              t <- typeIdentifier
              reservedOp "="
              u <- simtyp
              return $ ManifestTyp t u)
  <|> do reserved "type"
         t <- typeIdentifier
         return $ OpaqueTyp t
  <|> do reserved "val"
         x <- coreIdentifier
         colon
         v <- valtyp
         return $ ValSpec (ValId x) v

sigexp :: MLParser SigExp
sigexp = 
      do reserved "sig"
         elems <- many sigbodElem
         reserved "end"
         return $ SigBod elems
  <|> do x <- coreIdentifier
         option (SigVar (SigId x))
                (do reserved "where"
                    reserved "type"
                    t <- typeIdentifier
                    reservedOp "="
                    u <- simtyp
                    return $ SigWhere (SigVar (SigId x)) t u)
--
-- Program parser
--

sigbind :: MLParser SigBind
sigbind = 
    do reserved "signature"
       x <- coreIdentifier
       reservedOp "="
       sig <- sigexp
       return $ SigBind (SigId x) sig

rfun :: MLParser Rfun
rfun = 
  do reserved "functor"
     f <- coreIdentifier
     args <- parens (commaSep $ do x <- coreIdentifier
                                   colon 
                                   sig <- sigexp
                                   return (x, sig))
     let args' = if null args
                    then [("X__UNUSED__", SigBod [])]
                    else args
     colon
     sig <- sigexp
     reservedOp "="
     str <- sealedStrexp
     return $ Rfun (FunId f) (map (\ (x,y) -> (StrId x, y)) args') sig str

toplevelVal = 
    do reserved "val"
       x <- coreIdentifier
       reservedOp "="
       e <- valexp
       return $ ToplevelVal (ValId x) e

                                        
prog =
    do whiteSpace
       vals <- many toplevelVal
       sigs <- many sigbind
       reserved "rec"
       rfuns <- many rfun
       reserved "structure"
       symbol "Main"
       reservedOp "="
       reserved "struct"
       reserved "val"
       symbol "main"
       reservedOp "="
       e <- valexp
       reserved "end"
       eof
       let prog = Prog vals sigs (Rfuns rfuns) e
       return (removeTmpTyApps prog)

parseProg fname = 
    do src <- readFile fname
       return $ runParser prog emptyMLState fname src

parseString src = runParser prog emptyMLState "" src
    