module ML.Translation where

import Maybe ( fromJust )
import Control.Monad.State
import qualified Data.Map as Map
import qualified Data.Set as Set

import Phrac.Error
import Phrac.AbstractSyntax hiding ( ValId )
import qualified Phrac.AbstractSyntax as HS
import Phrac.Builtins
import Phrac.Pretty 
import Phrac.Symtab
import Phrac.Substitution
import Phrac.Unification
import Phrac.UniqIdents ( isTrue, isFalse )

import ML.AST
import ML.Identifiers
import ML.IdentifierManipulation

type MLValBind = (HS.ValId, ValExp)
type MLProg = Prog

--
-- Type translation
--

type TypeEnvironment = Map.Map TypeVarId SimTyp

emptyTypeEnvironment = Map.empty

data TTState = TTState { tt_tyenv :: TypeEnvironment,
                         tt_dictEnv :: Maybe [(ClassConstraint, ValExp)],
                         tt_argPairs :: Maybe [(StrId, SigExp)]
                       }

type TT = StateT TTState ST

errorPrefix = "translation HS->ML: "

runTT :: TypeEnvironment
      -> TT a        -- OV computation to run
      -> ST a
runTT tyenv tr = 
    evalStateT tr (TTState tyenv Nothing Nothing)

installTypeEnvironment :: TypeEnvironment -> TT ()
installTypeEnvironment sigma =
    modify (\s -> s { tt_tyenv = sigma })

deleteTypeEnvironment :: TT ()
deleteTypeEnvironment =
    modify (\s -> s { tt_tyenv = emptyTypeEnvironment })

installDictEnv :: [(ClassConstraint, ValExp)] -> TT ()
installDictEnv x =
    modify (\s -> s { tt_dictEnv = Just x })

deleteDictEnv :: TT ()
deleteDictEnv = 
    modify (\s -> s { tt_dictEnv = Nothing })

getDictEnv :: TT [(ClassConstraint, ValExp)]
getDictEnv =
    do x <- gets tt_dictEnv
       case x of
         Just y -> return y
         Nothing -> panic "no dictEnv available"

installArgPairs :: [(StrId, SigExp)] -> TT ()
installArgPairs x =
    modify (\s -> s { tt_argPairs = Just x })

deleteArgPairs :: TT ()
deleteArgPairs = 
    modify (\s -> s { tt_argPairs = Nothing })

getArgPairs :: TT [(StrId, SigExp)]
getArgPairs =
    do x <- gets tt_argPairs
       case x of
         Just y -> return y
         Nothing -> panic "no argPairs available"
{-
**PROBLEM** 
  We cannot guess the correct \Sigma' in rule (let)^t.

**SOLUTION**
  Tiny-HS type variables that are simply mapped to the corresponding
  Tiny-ML+ simple type variables are _not_ put into the constraint
  environment.
-}
lookupSimTyp :: TypeVarId -> TT SimTyp
lookupSimTyp a = 
    do tyenv <- gets tt_tyenv
       case Map.lookup a tyenv of
         Just u -> return u
         Nothing -> return $ SimTypTyvar (typVar2simTypVar a)

translateTypeScheme :: TypeScheme -> TT ValTyp
translateTypeScheme (TypeScheme qs rho) = 
    do u <- translateQualType rho
       let qs' = map typVar2simTypVar qs
       return $ ValTyp qs' u

translateQualType :: QualifiedType -> TT SimTyp
translateQualType (QualifiedType [] tau) = translateMonoType tau
translateQualType (QualifiedType (CC c : cs) tau) = 
    do sig <- sigForConstraint c
       u' <- translateQualType (QualifiedType cs tau)
       return $ tyFun [PackedStr sig, u']

translateQualType _ = 
    panic "translation HS->ML: types with equality constraints not supported"

translateMonoType :: Type -> TT SimTyp
translateMonoType tau = 
    do u <- trans tau
       return $ removeTmpTyApps u
    where trans (TyVar a) = lookupSimTyp a
          trans (TyApp t1 t2) = 
              do u1 <- trans t1
                 u2 <- trans t2
                 return $ TmpTyApp u1 u2
          trans (TyConstruct tid) = return $
              case () of 
               _| isFunTypeConstructor tid -> mkTycon "->"
                | isListTypeConstructor tid -> mkTycon "list"
                | isIntTypeConstructor tid -> mkTycon "int"
                | isBoolTypeConstructor tid -> mkTycon "bool"
                | otherwise -> mkTycon (showPpr tid)
--                | otherwise -> panic ("translation HS->ML: unknown type constructor: "
--                                      ++ showPpr tid)
          trans (TyAssoc _) = 
              panic ("translation HS->ML: associated type syonyms not supported")
          mkTycon s = SimTypTycon (TyconId s) [] 

--        
-- Expression translation
--

transOp :: Op -> [ValExp] -> ValExp
transOp op args = 
    if length args /= 2
       then panic (errorPrefix ++ "unsupported operator: " ++ showPpr op)
       else let op' = case op of
                        IntAddOp -> Add
                        IntSubOp -> Sub
                        IntMulOp -> Mult
  	                IntQuotOp -> Div
  	                IntGtOp -> Gt
  	                IntGeOp -> Geq
  	                IntLtOp -> Lt
  	                IntLeOp -> Leq
  	                IntEqOp -> Eq
                        IntNeOp -> Neq
                        _ -> panic (errorPrefix ++ "unsupported operator: " 
                                    ++ showPpr op)
                in BinExp op' (args!!0) (args!!1)


transCon :: DataId -> [ValExp] -> ValExp
transCon di es =
    case () of
      _| isNil di -> NilExp
       | isCons di -> case es of
                        [e1, e2] -> ConsExp e1 e2
                        l -> panic ("type error in list expression, arguments: " ++
                                    showPprTML es)
       | isTrue di -> CodeExp "true"
       | isFalse di -> CodeExp "false"
       | otherwise -> panic ("unsupported data constructor: " ++ showPpr di)


--
-- Instance translation
--

prepareInstanceTranslation :: ClassInst -> TT ()
prepareInstanceTranslation ci =
    do -- preparations
       let constrs = inst_constraints ci
       xs <- lift $ mapM (\_ -> freshStrId) constrs
       y <- lift $ freshStrId
       let (argPairs', sigma') = collectArgsAndTypeEnv [] xs constrs
           allFree = freeTypeVars (inst_types ci)
           constrainedFree = freeTypeVars (inst_constraints ci)
           unconstrainedFree = Set.toList (allFree `Set.difference` constrainedFree)
           argPairs = argPairs' ++ [(y, SigBod (map (OpaqueTyp . typVar2typId)
                                                unconstrainedFree))]
           sigma = foldl (\ m a -> Map.insert a (QModTyp y (typVar2typId a)) m)
                     sigma' unconstrainedFree
           dictEnv = map (\ (x,CC c) -> (c, PackExp (StrVar x) 
                                            (SigWhere (fromJust $ lookup x argPairs)
                                             typeId (QModTyp x typeId))))
                         (zip xs constrs)
       installTypeEnvironment sigma
       installDictEnv dictEnv
       installArgPairs argPairs
       return ()
    where collectArgsAndTypeEnv :: [(TypeVarId, StrId)] -- list of type variables
                                                        -- already represented by a
                                                        -- certain argument
                                -> [StrId]
                                -> [Constraint]
                                -> ([(StrId, SigExp)], TypeEnvironment)
          collectArgsAndTypeEnv _ _ [] = ([], Map.empty)
          collectArgsAndTypeEnv used (x:xs) (CC (ClassConstraint cid [TyVar a]) 
                                             : rest) =
              case lookup a used of
                     Nothing -> let (args, sigma) = collectArgsAndTypeEnv
                                                      ((a,x):used) xs rest
                                    in ((x, SigVar $ classId2sigId cid) : args,
                                        Map.insert a (QModTyp x typeId) sigma)
                     Just y -> let (args, sigma) = collectArgsAndTypeEnv used xs rest
                                   in ((x, SigWhere (SigVar $ classId2sigId cid)
                                             typeId (QModTyp y typeId)) : args,
                                       sigma)

translateInstance :: ClassInst -> [MLValBind] -> TT Rfun
translateInstance ci mlBinds = 
    do display ("translating instance `" ++ showInstHead ci ++ "'")
       -- preparations
       dictEnv <- getDictEnv
       argPairs <- getArgPairs
       let t = case inst_types ci of
                 [t] -> t
                 _ -> panic (errorPrefix ++ 
                             "multi-parameter type classes not supported")
       -- handle superclasses
       display ("translating superclasses")
       let cid = inst_class ci
       cdec <- lift $ findClass cid
       superClasses <- lift $ findSuperClasses cid
       let supers = map (\ c -> ClassConstraint c [t]) superClasses
       superExps <- mapM (\c -> runET dictEnv (getDictionaryExp c)) supers
       let superBinds = map (\ (c,e) -> (superId c, e)) (zip supers superExps)
           superValdefs = map (\ (x, e) -> ValDef x Nothing e) superBinds
       display ("value definitions for superclasses: " ++
                show (vcat (map pprTML superValdefs)))
       -- handle methods
       display ("translating methods")
       msigs' <- lift $ 
            mapM (\ (m,_) -> findMethodSig cid m >>= \ (_, sc) -> return sc)
                 mlBinds
       let subst = substFromAssocs (zip (class_params cdec) (inst_types ci))
           msigs = applySubst subst msigs'
       sigsML <- mapM translateTypeScheme msigs
       let methodValdefs = map (\ ((x, e), v) -> 
                                ValDef (methodId2valId x) (Just v) (rwMVal dictEnv e))
                               (zip mlBinds sigsML)
       display ("value definitions for methods: " ++
                show (vcat (map pprTML methodValdefs)))
       f <- lift $ getFunId ci
       u <- translateMonoType t
       let ressig = SigWhere (SigVar $ classId2sigId cid) typeId u
           body = StrBod $ TypDef typeId u : (superValdefs ++ methodValdefs)
           res = Rfun f argPairs ressig body
       display ("finished translating instance `" ++ showInstHead ci ++ "': "
               ++ showPprTML res)
       deleteTypeEnvironment 
       deleteDictEnv
       deleteArgPairs
       return res
   where rwMVal :: [(ClassConstraint, ValExp)] -> ValExp -> ValExp
         -- applies a method translation to the evidence expressions for the
         -- instance constraints
         rwMVal l e =
             let args = map snd l
                 in foldl AppExp e args

--
-- Class translation
--

translateClass :: ClassDec -> ST SigBind
translateClass (ClassDec i [a] supers [] vsigs _) = 
    do let sigma = Map.singleton a (ModTyp typeId)
           methodIds = map (methodId2valId . fst) vsigs
       methodSigs <- runTT sigma $ mapM translateTypeScheme (map snd vsigs)
       let superIds = map cc_name supers
           superSigs = map (\c -> SigWhere (SigVar $ classId2sigId c) typeId
                                    (ModTyp typeId)) superIds
           valSpecs = map (uncurry ValSpec) (zip methodIds methodSigs) ++
                      map (uncurry ValSpec) (zip (map superId' superIds) 
                                             (map (\s -> ValTyp [] $ PackedStr s)
                                              superSigs))
           sigexp = SigBod (OpaqueTyp typeId : valSpecs)
       return $ SigBind (classId2sigId i) sigexp


--
-- Entailment translation
--

data ETState = ETState { -- mapping from class constraints to dictionary 
                         -- variables providing evidence for the constraint
                         et_dictEnv :: DictEnv 
                       }

type ET = StateT ETState TT

runET :: [(ClassConstraint, ValExp)]   -- mapping from class constraints to
                                       -- dictionary variables for the binding
                                       -- being translated
      -> ET a        -- ET computation to run
      -> TT a
runET pairs et = 
    do env <- lift $ buildDictEnv pairs
       evalStateT et (ETState env)

runETTT pairs sigma et = 
    runTT sigma (runET pairs et)



-- The dictionary environment.
--
-- Maps class constraints to triples with the
-- expression holding the dictionary for the constraints,
-- the class of the dictionary
-- and the path from the class of the dictionary to the class of the constraint.
-- If there is a mapping from `C a1 .. aN' to
-- some tuple, there must be mappings for all super classes of `C'
-- to the same tuple. Use the `buildDictEnv' function to maintain this
-- invariant.
-- 
-- Example:
-- Consider C3 is a subclass of C2 which is a subclass of C1,
-- and there is a dictionary variable d for C3 t.
-- Then the dictionary environment looks like this:
-- { C3 t -> (e, C3, [])
--   C2 t -> (e, C3, [C2])
--   C1 t -> (e, C3, [C2, C1]) }
type DictEnv = Map.Map ClassConstraint (ValExp, ClassId, ClassPath)
emptyDictEnv = Map.empty

buildDictEnv :: [(ClassConstraint, ValExp)] -> ST DictEnv
buildDictEnv pairs = foldM ins emptyDictEnv pairs
    where ins env (cc@(ClassConstraint c t), e) = 
              do supers <- findSuperClassPath c
                 let constrs = cc : map (\p -> ClassConstraint (last p) t) supers
                     paths   = [] : supers 
                 return (foldl (\m (c',p) -> Map.insert c' (e, c, p) m) env 
                         (zip constrs paths))


-- Returns a triple with the instance declaration which fulfills the given 
-- constraint, the unconstrained type variables of the instance declaration
-- mapped to their instantiations, and
-- the instantiated class constraints in the instance context.
-- If there is no instance declaration fulfilling the constraint,
-- `Nothing' is returned. This can happen with deferred constraints.
getInstance :: ClassConstraint -> ST (Maybe (ClassInst, 
                                             [(TypeVarId, Type)],
                                             [ClassConstraint]))
getInstance c@(ClassConstraint i _) =
    do insts <- findInstances i
       candidates <- sequence [tryInst it | it <- insts]
       case msum candidates of
         Nothing -> 
             return Nothing
         Just (ci, tyvarMapping, goals) ->
             let (cs, _) = splitConstraints goals
                 in return $ Just (ci, tyvarMapping, cs)
     where tryInst ci =
              do u <- matchConstraint (inst_class ci, inst_types ci) c
                 case u of
                   Nothing -> return Nothing
                   Just u' -> 
                       let allFree = freeTypeVars (inst_types ci)
                           constrainedFree = freeTypeVars (inst_constraints ci)
                           unconstrainedFree = Set.toList (allFree `Set.difference`
                                                           constrainedFree)
                           tyvarMapping = map (\a -> (a, applySubst u' (TyVar a)))
                                          unconstrainedFree
                       in return $ Just (ci, tyvarMapping,
                                         map (applySubst u') (inst_constraints ci))


{-
  Returns a expression evaluating to a dictionary which provides evidence
  for the given constraint.
-}
getDictionaryExp :: ClassConstraint -> ET ValExp
getDictionaryExp c@(ClassConstraint _ (t:[])) = 
    do env <- gets et_dictEnv
       case Map.lookup c env of
         -- Need to check if `id' represents a subclass of the given constraint.
         -- If so, I must generate code to extract the right superclass
         -- dictionary (see rule super_entail)
         Just (e, c', p) -> 
             lift $ accessSuperDict e (ClassConstraint c' [t]) p
         Nothing ->
           do inst <- lift $ lift $ getInstance c
              case inst of
                Nothing -> 
                    return $ OverloadingML (Placeholder c) -- deferred constraint
                Just (ci, unconstrained, cs) ->
                    -- rule inst_entail
                    do dicts <- mapM getDictionaryExp cs
                       xs <- lift $ lift $ mapM (\_ -> freshStrId) dicts
                       f <- lift $ lift $ getFunId ci
                       sig <- lift $ sigForConstraint c                           
                       y <- lift $ strExpForUnconstrainedFreeTypeVariables 
                                     unconstrained
                       let fapp = FunctorApp f (map StrVar xs ++ [y])
                           packExp = PackExp fapp sig
                       openExp <- mkOpenExp (zip3 dicts xs cs) packExp
                       return openExp
    where mkOpenExp [] e = return e
          mkOpenExp ( (de, x, cc) : rest ) e = 
              do sig <- lift $ sigForConstraint cc
                 e' <- mkOpenExp rest e
                 return $ OpenExp de x sig e'

getDictionaryExp _ = 
    panic ("translation ML->HS: multi-parameter type classes not supported")


accessSuperDict :: ValExp          -- the expression containing the 
                                   -- subclass dictionary
                -> ClassConstraint -- the "type" of the subclass dictionary
                                   -- expression
                -> ClassPath       -- the class path, _not_ containing
                                   -- the "type" of the subclass dictionary
                                   -- expression
                -> TT ValExp
accessSuperDict e _ [] = return e
accessSuperDict e cc@(ClassConstraint _ (t:[])) (c:cs) = 
    do x <- lift $ freshStrId
       sig <- sigForConstraint cc
       let e' = OpenExp e x sig (superVal x c)
       accessSuperDict e' (ClassConstraint c [t]) cs
accessSuperDict _ _ _ = 
    panic ("translation ML->HS: multi-parameter type classes not supported")



--
-- Utilities
--

sigForConstraint :: ClassConstraint -> TT SigExp
sigForConstraint (ClassConstraint c [t]) = 
    do u <- translateMonoType t
       let id = classId2sigId c
       return $ SigWhere (SigVar id) typeId u
sigForConstraint (ClassConstraint c _) = 
    panic (errorPrefix ++ "multi-parameter type classes not supported")

superVal :: StrId -> ClassId -> ValExp
superVal x c = QModVarExp x (superId' c)

methodVal :: StrId -> HS.ValId -> ValExp
methodVal x m = QModVarExp x (methodId2valId m)

strExpForUnconstrainedFreeTypeVariables :: [(TypeVarId, Type)] -> TT StrExp
strExpForUnconstrainedFreeTypeVariables l =
    do l' <- mapM (\ (a,t) -> translateMonoType t >>= \u -> return (a,u)) l
       let tdefs = map (\(a,u) -> TypDef (typVar2typId a) u) l'
       return $ StrBod tdefs

mkLam :: [CoreId] -> ValExp -> ValExp
mkLam [] e = e
mkLam (c:cs) e = LamExp c (mkLam cs e)


