{-# OPTIONS_GHC -fglasgow-exts #-}

-- 
-- Copyright (C) 2005 Stefan Wehr - http://www.stefanwehr.de
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License as
-- published by the Free Software Foundation; either version 2 of
-- the License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-- General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.
-- 

module Phrac.Overloading (

  resolveOverloading, 
  dumpDictExprs

) where

import qualified Data.Map as Map
import qualified Data.Set as Set
import Control.Monad.State

import Phrac.Symtab
import Phrac.AbstractSyntax hiding ( ValId )
import Phrac.Error
import Phrac.Unification ( matchConstraint )
import Phrac.Pretty 
import Phrac.Substitution

import ML.AST
import ML.Identifiers
import ML.Translation
import ML.IdentifierManipulation

{-

The dictionary resolution algorithm is based on "John Peterson and 
Mark Jones. Implementing Type Classes. PLDI '93". 

-}



data OVState = OVState { -- the current substitution
                         ov_subst  :: TypeSubst,
                         -- context information for recursivly defined variables
                         ov_recMap :: RecMap
                       }

type OV = StateT OVState ET

runOV :: TypeSubst   -- current substitution
      -> RecMap      -- context for recursivly defined variables
      -> [(ClassConstraint, CoreId)]   -- mapping from class constraints to
                                       -- dictionary variables for the binding
                                       -- being translated
      -> OV a        -- OV computation to run
      -> TT a
runOV subst rm pairs ov = 
    do let pairs' = map (\ (c,i) -> (c, CoreVarExp i)) pairs
       let et = evalStateT ov (OVState subst rm)
       runET pairs' et

type RecMap = Map.Map CoreId [ClassConstraint]
emptyRecMap = Map.empty


resolveExp :: ValExp -> OV ValExp

resolveExp (OverloadingML (Placeholder c)) = 
    do s <- gets ov_subst
       lift $ getDictionaryExp (applySubst s c)

resolveExp (OverloadingML (MethodPlaceholder cid mid ts)) = 
    do s <- gets ov_subst
       let ts' = applySubst s ts
           cc = (ClassConstraint cid ts')
       sig <- lift $ lift $ sigForConstraint cc
       dexp <- lift $ getDictionaryExp cc
       x <- lift $ lift $ lift $ freshStrId
       return $ OpenExp dexp x sig (methodVal x mid)

resolveExp (OverloadingML o) = 
    panic ("unexpected Overloading value during overloading resolution: " ++
           show o)

resolveExp e@(CoreVarExp _) = return e
resolveExp e@(IntExp _) = return e
resolveExp (BinExp op e1 e2) = 
    do e1' <- resolveExp e1
       e2' <- resolveExp e2
       return $ BinExp op e1' e2'
resolveExp (LamExp c e) = 
    do e' <- resolveExp e
       return $ LamExp c e'
resolveExp (AppExp e1 e2) = 
    do e1' <- resolveExp e1
       e2' <- resolveExp e2
       return $ AppExp e1' e2'
resolveExp (IfExp e1 e2 e3) = 
    do e1' <- resolveExp e1
       e2' <- resolveExp e2
       e3' <- resolveExp e3
       return $ IfExp e1' e2' e3'
resolveExp (LetExp c v e1 e2) =
    do e2' <- resolveExp e2
       return $ LetExp c v e1 e2'
resolveExp e@(ModVarExp _) = return e
resolveExp e@(QModVarExp _ _) = return e
resolveExp e@(PackExp _ _) = return e
resolveExp e@(OpenExp e1 x s e2) = 
    do e1' <- resolveExp e1
       e2' <- resolveExp e2
       return $ OpenExp e1' x s e2'
resolveExp (ListExp es) = 
    do es' <- mapM resolveExp es
       return $ ListExp es'
resolveExp NilExp = return NilExp
resolveExp (ConsExp e1 e2) = 
    do e1' <- resolveExp e1
       e2' <- resolveExp e2
       return $ ConsExp e1' e2'
resolveExp e@(CodeExp _) = return e

resolveOverloading :: TypeSubst     -- the current substitution
                   -> RecMap        -- constraints for recursivly defined vars
                   -> ValExp          -- the expression to transform
                   -> [ClassConstraint]  -- class context of the binding 
                   -> TT ValExp         -- the transformed expression
resolveOverloading subst recMap e cs =
    do debug ("resolveOverloading called with recMap=" ++ show (dumpMap recMap)
              ++ ", cs=" ++ showCommaListed cs)
       ds <- lift $ mapM (\_ -> freshCoreId) cs
       e' <- runOV subst recMap (zip cs ds) (resolveExp e)
       debug ("result of resolveOverloading: " ++ show e')
       return (mkLam ds e')


dumpDictExprs :: [((ClassId, [Type]), Exp)] -> Doc
dumpDictExprs [] = empty
dumpDictExprs (((cid, ts), e) : rest) = 
    let restDoc = dumpDictExprs rest
        doc = text "dict-" <> ppr cid <> text "<" <> pprList ", " ts 
              <> text ">:" $$ ppr e
        in doc $$ restDoc


