module Main where

import Control.Exception        ( catchDyn, evaluate )
import Control.Monad            ( when )
import System.Console.GetOpt
import System.IO                ( hFlush, stdin, stdout, stderr, hPutStrLn,
                                  hGetContents )
import System

import Phrac.Error
import Phrac.Constants
import Phrac.Pretty

import ML.TMLParser
import ML.Mosml


dump' :: [Flag] -> String -> IO ()
dump' fs s =
    if not (null s)
       then do hPutStrLn stdout s
               if ExitAfterDump `elem` fs
                  then exitWith ExitSuccess
                  else return ()
       else return ()

tmlmain flags fname = 
    do let on = \x -> elem x flags
       let dump p = dump' flags p
   
       case logLevel flags of
         Nothing -> return ()
         Just l -> enableDebug l
   
       display ("parsing " ++ fname)
       pres <- parseProg fname
       ast <- case pres of
                Right x -> return x
                Left err -> panic ("Parsing of source code failed!\n\n" ++
                                   show err)
   
       when (on DumpRawParse) $ dump (show ast)
       when (on DumpMosml) $ dump (show.ppr $ ast)

       display ("running Moscow ML")
       answ <- runMosml ast (mosmlCmd flags)

       return answ
   
myCatch :: [Flag] -> IO a -> IO (Either String a)
myCatch flags io = catchDyn (io >>= \p -> return $ Right p) handler 
    where handler dyn =
              do let msg = show (dyn :: PException)
                 return $ Left msg

main = 
    do args <- getArgs
       prog <- getProgName
       (flags,fname) <- parseArgs args prog

       res <- myCatch flags (tmlmain flags fname)
       case res of
         Right s -> putStr s
         Left s -> putStrLn ("ERROR\n" ++ s)


-- ---------------------------------------------------------------------
-- Handling command line flags

logLevel :: [Flag] -> Maybe DebugLevel
logLevel [] = Nothing
logLevel (LogLevel l:_) = Just l
logLevel (_:rest) = logLevel rest

mosmlCmd :: [Flag] -> MosmlCmd
mosmlCmd [] = "mosml"
mosmlCmd (MosmlCmdFlag c : _) = c
mosmlCmd (_:rest) = mosmlCmd rest

data Flag   
    = Version
    | LogLevel DebugLevel
    | Help 
    | DumpMosml
    | DumpRawParse
    | ExitAfterDump
    | MosmlCmdFlag String
  deriving Eq

--
-- associate concrete flags with their abstract representation
--
opts :: [OptDescr Flag]
opts = 
    [ Option ['v'] ["version"] (NoArg Version) "version string"
    , Option ['?'] ["help"] (NoArg Help) "this message"
    , Option []    ["debug"] (NoArg (LogLevel Verbose)) "enable debug messages"
    , Option []    ["verbose"] (NoArg (LogLevel Verbose)) 
            "be verbose (but not as much as with `--debug')"
    , Option []    ["exit-after-dump"] (NoArg ExitAfterDump) 
            "terminate the program after dumping"
    , Option []    ["dump-mosml"] (NoArg DumpMosml)
            "dump Moscow ML source code"
    , Option []    ["dump-ast"] (NoArg DumpRawParse)
            "dump raw abstract syntax"
    , Option []    ["mosml"] (ReqArg MosmlCmdFlag "PATH")
            "path to `mosml', the interactive system of Moscow ML"
    ]

--
-- parse the args.
--
-- we also short circuit here, for the known cases of --version or --help
--
parseArgs :: [String] -> String -> IO ([Flag],FilePath)
parseArgs argv prog = 
    case (getOpt Permute opts argv) of
      (flags,fs,[]) -> case () of { _ 
         | Version `elem` flags -> do putStrLn phracVersion
                                      exitWith ExitSuccess
         | Help `elem` flags -> do putStrLn usage
                                   exitWith ExitSuccess
         | otherwise -> do if length fs == 1
                              then return (flags,fs!!0)
                              else do putStrLn usage
                                      exitFailure
      }
      (_,_,err) -> do putStrLn (concat err ++ usage)
                      exitFailure
    where 
    usage = usageInfo ("Usage: " ++ prog ++ " [OPTION...] file") opts

